from .utils import _Attribute, _Comment, _parse
from .dom import COMMENT, DOCUMENT_TYPE, TEXT, ELEMENT, FRAGMENT
from .dom import Node, Comment, DocumentType, Text, Element, Fragment, parse, unsafe, _clone


_parsed = {}
_listeners = []

# T = (t'').__class__
# Let's just use the imported class
try:
  # Temporary hack because pytest-playwright has to run under 3.13
  # in CI due to greenlet not compiling there under 3.14.
  from string.templatelib import Template
except ImportError:
  pass

def _util(svg):
  def fn(t):
    if not isinstance(t, Template):
      raise ValueError('Argument is not a Template instance')

    strings = t.strings

    values = [entry.value for entry in t.interpolations]

    length = len(values)

    if not strings in _parsed:
      _parsed[strings] = _parse(strings, length, svg)

    content, updates = _parsed[strings]

    node = _clone(content)
    changes = []
    path = None
    child = None
    i = 0

    for update in updates:
      if path != update.path:
        path = update.path
        child = node
        for index in path:
          child = child['children'][index]

      if isinstance(update.value, _Attribute):
        changes.append(update.value(child, _listeners))
      elif isinstance(update.value, _Comment):
        changes.append(update.value(child))
      else:
        changes.append(update.value(child, changes))

    for i in range(length):
      changes[i](values[i])

    for i in range(len(changes) - 1, length - 1, -1):
      changes[i]()

    return node

  return fn


def render(where, what):
  result = where(what() if callable(what) else what, _listeners)
  _listeners.clear()
  return result

html = _util(False)
svg = _util(True)


__all__ = [
  'render', 'html', 'svg', 'parse', 'unsafe',
  'Node', 'Comment', 'DocumentType', 'Text', 'Element', 'Fragment',
  'COMMENT', 'DOCUMENT_TYPE', 'TEXT', 'ELEMENT', 'FRAGMENT',
]



"""Subcomponents."""

from tdom import html

title = "My Todos"


def Todo(label):
    """An individual to do component."""
    return html(t"<li>{label}</li>")


def TodoList(todos):
    """A to do list component."""
    # Pass an argument, not the normal props style
    return html(t"<ul>{[Todo(label) for label in todos]}</ul>")


def main():
    """Main entry point."""
    todos = ["first"]
    return html(
        t"""
      <h1>{title}</h1>
      <{TodoList} todos={todos} />
    """
    )



"""
Port the code from viewdom.examples.static_string.
"""

import pytest

from tdom import html


def test_string_literal():
    """Simplest form of templating: just a string, no tags, no attributes"""
    result = html(t"Hello World")
    assert str(result) == "Hello World"


def test_simple_render():
    """Same thing, but in a `<div> with attributes`."""
    result = html(t'<div title="Greeting">Hello World</div>')
    assert str(result) == '<div title="Greeting">Hello World</div>'


def test_simple_interpolation():
    """Simplest rendering of a value in the scope."""
    name = "World"
    result = html(t"<div>Hello {name}</div>")
    assert str(result) == "<div>Hello World</div>"


def test_attribute_value_expression():
    """Pass in a Python symbol as part of the template, inside curly braces."""
    klass = "container1"
    result = html(t"<div class={klass}>Hello World</div>")
    assert str(result) == '<div class="container1">Hello World</div>'


def test_expressions_in_attribute_value():
    """Simple Python expression cannot be used inside an attribute value."""
    try:
        result = html(t'<div class="container{1}">Hello World</div>')
        raise Exception("Should not happen")
    except ValueError as e:
        assert str(e) == "0 updates found, expected 1"


def test_child_nodes():
    """Nested markup shows up as nodes."""
    result = html(t"<div>Hello <span>World<em>!</em></span></div>")
    assert str(result) == "<div>Hello <span>World<em>!</em></span></div>"


def test_doctype():
    """Sometimes it is hard to get a DOCTYPE in to the resulting output."""
    result = html(t"<!DOCTYPE html>\n<div>Hello World</div>")
    assert str(result) == "<!DOCTYPE html><div>Hello World</div>"


def test_reducing_boolean():
    """collapse truthy-y values into simplified HTML attributes."""
    result = html(t"<div editable={True}>Hello World</div>")
    assert str(result) == "<div editable>Hello World</div>"



"""Child nodes become part of the structure."""
from tdom import html


def main():
    """Main entry point."""
    structure = html(t"<div>Hello <span>World<em>!</em></span></div>")
    return structure



import sys
from html import escape
from random import random


_IS_MICRO_PYTHON = "MicroPython" in sys.version


_prefix = 't🐍' + str(random())[2:5]
_data = f'<!--{_prefix}-->'



ELEMENT = 1
# ATTRIBUTE = 2
TEXT = 3
# CDATA = 4
# ENTITY_REFERENCE = 5
# ENTITY = 6
# PROCESSING_INSTRUCTION = 7
COMMENT = 8
# DOCUMENT = 9
DOCUMENT_TYPE = 10
FRAGMENT = 11
# NOTATION = 12



TEXT_ELEMENTS = (
  'plaintext',
  'script',
  'style',
  'textarea',
  'title',
  'xmp',
)


VOID_ELEMENTS = (
  'area',
  'base',
  'br',
  'col',
  'embed',
  'hr',
  'img',
  'input',
  'keygen',
  'link',
  'menuitem',
  'meta',
  'param',
  'source',
  'track',
  'wbr',
)


class Node(dict):
  def __init__(self, **kwargs):
    super().__init__(type=self.type, **kwargs)
    self.parent = None

  def __getattr__(self, name):
    return self[name] if name in self else None


class Comment(Node):
  type = COMMENT

  def __init__(self, data):
    super().__init__(data=data)

  def __str__(self):
    return f'<!--{str(self['data'])}-->'


class DocumentType(Node):
  type = DOCUMENT_TYPE

  def __init__(self, data):
    super().__init__(data=data)

  def __str__(self):
    return f'<!{self['data']}>'


class Text(Node):
  type = TEXT

  def __init__(self, data):
    super().__init__(data=data)

  def __str__(self):
    data = self['data']
    return data if isinstance(data, Unsafe) else escape(str(data))


class Element(Node):
  type = ELEMENT

  def __init__(self, name, xml=False):
    super().__init__(name=name, xml=xml, props={}, children=[])

  def __str__(self):
    xml = self['xml']
    name = self['name']
    html = f'<{name}'
    for key, value in self['props'].items():
      if value != None:
        if isinstance(value, bool):
          if value:
            html += f' {key}=""' if xml else f' {key}'
        else:
          html += f' {key}="{escape(str(value))}"'
    if len(self['children']) > 0:
      html += '>'
      just_text = not xml and name.lower() in TEXT_ELEMENTS
      for child in self['children']:
        html += child['data'] if just_text else str(child)
      html += f'</{name}>'
    elif xml:
      html += ' />'
    else:
      html += '>'
      if not name.lower() in VOID_ELEMENTS:
        html += '</' + name + '>'
    return html


class Fragment(Node):
  type = FRAGMENT

  def __init__(self):
    super().__init__(children=[])

  def __str__(self):
    return ''.join(str(child) for child in self['children'])



def _append(parent, node):
  parent['children'].append(node)
  node.parent = parent


def _appendChildren(parent, nodes, clone=False):
  children = parent['children']
  for node in nodes:
    if clone:
      node = _clone(node)
    children.append(node)
    node.parent = parent


def _clone(node):
  type = node['type']
  if type == FRAGMENT:
    fragment = Fragment()
    _appendChildren(fragment, node['children'], True)
    return fragment
  if type == ELEMENT:
    element = Element(node['name'], node['xml'])
    element['props'] = node['props'].copy()
    _appendChildren(element, node['children'], True)
    return element
  if type == TEXT:
    return Text(node['data'])
  if type == COMMENT:
    return Comment(node['data'])
  if type == DOCUMENT_TYPE:
    return DocumentType(node['data'])


def _replaceWith(current, node):
    parent = current.parent
    children = parent['children']
    children[children.index(current)] = node
    node.parent = parent
    current.parent = None



if _IS_MICRO_PYTHON:
  import re

  ATTRIBUTES = re.compile(r'([^\s=]+)(=(([\'"])[\s\S]*?\4|\S+))?')
  NAME_CHAR = re.compile(r'[^/>\s]')


  class Unsafe(str):
    def __init__(self, *args, **kwargs):
      super().__init__(*args, **kwargs)


  def _attributes(props, attrs):
    while match := ATTRIBUTES.match(attrs.strip()):
      key = match.group(1)
      equal = match.group(2)
      if equal:
        value = match.group(3)
        # somehow MicroPython doesn't support match.group(4) for quoted values
        if value and value[0] == value[-1] and (value[0] == '"' or value[0] == "'"):
          value = value[1:-1]
        props[key] = value
        i = attrs.index(equal) + len(equal)
      else:
        props[key] = True
        i = attrs.index(key) + len(key)

      attrs = attrs[i:]

  def _text(node, content, ts, te):
    if ts < te:
      data = content[ts:te]
      if data.strip():
        _append(node, Text(data))

  def parse(content, xml=False):
    node = Fragment()
    length = len(content)
    i = 0
    ts = 0
    te = 0
    while i < length:
      i += 1
      if content[i - 1] == '<':

        if content[i] == '/':
          _text(node, content, ts, te)
          ts = te = i = content.index('>', i + 1) + 1
          node = node.parent
          continue

        if content[i] == '!':
          _text(node, content, ts, te)
          if content[i+1:i+3] == '--':
            j = content.index('-->', i + 3)
            data = content[i+3:j]
            if not (data.startswith('#') and data.endswith('#')):
              _append(node, Comment(data))
            i = j + 3
          else:
            j = content.index('>', i + 1)
            _append(node, DocumentType(content[i+1:j]))
            i = j + 1
          ts = te = i
          continue

        j = i
        while j < length and NAME_CHAR.match(content[j]):
          j += 1

        if i < j:
          _text(node, content, ts, te)
          name = content[i:j]
          element = Element(name, xml)
          _append(node, element)

          not_void = not xml and not name.lower() in VOID_ELEMENTS
          if not_void:
            node = element

          i = j
          j = content.index('>', i)
          closing = content[j - 1] == '/'
          if closing:
            j -= 1
          if i < j:
            _attributes(node['props'], content[i:j])
          if closing:
            j += 1
            if not_void:
              node = node.parent
          ts = te = i = j + 1
          continue

      te += 1

    _text(node, content, ts, te)
    return node

else:
  from html.parser import HTMLParser


  class Unsafe(str):
    def __new__(cls, value, *args, **kwargs):   
      return super(Unsafe, cls).__new__(cls, value)


  class DOMParser(HTMLParser):
    def __init__(self, xml=False):
      super().__init__()
      self.xml = xml
      self.node = Fragment()

    def handle_starttag(self, tag, attrs):
      element = Element(tag, self.xml)
      _append(self.node, element)

      if not self.xml and not tag.lower() in VOID_ELEMENTS:
        self.node = element

      props = element['props']
      for name, value in attrs:
        props[name] = value

    def handle_endtag(self, tag):
      if not self.xml and not tag.lower() in VOID_ELEMENTS:
        parent = self.node.parent
        if parent:
          self.node = parent

    def handle_data(self, data):
      # this is needed to handle sparse interpolations
      # within <style> or <script> tags where this parser
      # won't allow children nodes and it passes all as data
      text = data.split(_data)
      for i in range(len(text) - 1):
        # empty nodes are ignored
        if len(text[i].strip()) > 0:
          _append(self.node, Text(text[i]))
        # the comment node though is needed to handle updates
        _append(self.node, Comment(_prefix))

      # same applies for the last node
      if len(text[-1].strip()) > 0:
        _append(self.node, Text(text[-1]))

    def handle_comment(self, data):
      if data == '/':
        self.handle_endtag(self.node['name'])
      elif not (data.startswith('#') and data.endswith('#')):
        _append(self.node, Comment(data))

    def handle_decl(self, data):
      _append(self.node, DocumentType(data))

    def unknown_decl(self, data):
      raise Exception(f'Unknown declaration: {data}')


  def parse(content, xml=False):
    parser = DOMParser(xml)
    parser.feed(content)
    return parser.node


def unsafe(value):
  return Unsafe(value)



"""Pass a component as a prop value."""

from tdom import html


def DefaultHeading():
    """The default heading."""
    return html(t"<h1>Default Heading</h1>")


def Body(heading):
    """The body which renders the heading."""
    return html(t"<body><{heading} /></body>")


def main():
    """Main entry point."""
    result = html(t"<{Body} heading={DefaultHeading} />")
    return result



from .dom import VOID_ELEMENTS, _prefix
import re


_elements = re.compile(
  '<(\x01|[a-zA-Z0-9]+[a-zA-Z0-9:._-]*)([^>]*?)(/?)>',
)

_attributes = re.compile(
  r'([^\s\\>\'"=]+)\s*=\s*([\'"]?)' + '\x01',
)

_holes = re.compile(
  '[\x01\x02]',
)


def _as_attribute(match):
  return f'\x02={match.group(2)}{match.group(1)}'


def _as_closing(name, xml, self_closing):
  if len(self_closing) > 0:
    if xml or name.lower() in VOID_ELEMENTS:
      return ' /'
    else:
      return f'></{name}'
  return ''

# \x01 Node.ELEMENT_NODE
# \x02 Node.ATTRIBUTE_NODE

def _instrument(template, xml):
  def pin(match):
    name = match.group(1)
    if name == '\x01':
      name = _prefix
    attrs = match.group(2)
    self_closing = match.group(3)
    return f'<{
      name
    }{
      re.sub(_attributes, _as_attribute, attrs).rstrip()
    }{
      _as_closing(name, xml, self_closing)
    }>'

  def point():
    nonlocal i
    i += 1
    return _prefix + str(i - 1)

  i = 0
  return re.sub(
    _holes,
    lambda match: f'<!--{_prefix}-->' if match.group(0) == '\x01' else point(),
    re.sub(
      _elements,
      pin,
      '\x01'.join(template).strip().replace('</>', '<//>')
    )
  )



"""Call a function from inside a template expression."""

from tdom import html


def make_bigly(name: str) -> str:
    """A function returning a string, rather than a component."""
    return f"BIGLY: {name.upper()}"


def main():
    """Main entry point."""
    name = "tdom"
    result = html(t"<div>Hello {make_bigly(name)}</div>")
    return result



"""Get a variable from an import."""

from tdom import html

name = "World"


def Hello():
    """A simple hello component."""
    return html(t"<div>Hello {name}</div>")


def main():
    """Main entry point."""
    result = Hello()
    return result



"""Boolean attribute values are reduced during rendering."""

from tdom import html


def main():
    """Main entry point."""
    result = html(t"<div editable={True}>Hello World</div>")
    return result



"""Test an example."""

import pytest

from . import main


# TODO Andrea this test should pass
@pytest.mark.skip(reason="Broken in latest generator support")
def test_main() -> None:
    """Ensure the demo matches expected."""
    assert str(main()) == "<ul><li>First</li><li>Second</li></ul>"



"""Test an example."""
from . import main
from tdom import Element

def test_main() -> None:
    """Ensure the demo matches expected."""
    result: Element = main()
    assert result.props == {}
    # Test first child
    assert result.children[0] == {'type': 3, 'data': 'Hello '}
    assert str(result) == '<div>Hello <span>World<em>!</em></span></div>'



"""Use normal Python syntax for conditional rendering in a template."""

from tdom import html


def main():
    """Main entry point."""
    message = "Say Howdy"
    not_message = "So Sad"
    show_message = True
    result = html(
        t"""
        <h1>Show?</h1>
        {message if show_message else not_message}
    """
    )
    return result
